// shopbot.cpp 
// ackmed@gotwalls.com


#include "shopbot.h"
#include "miscstructs.h"
#include "item.h"
#include "Weapons.h"
#include "packethelper.h"
#include "waypoint.h"
#include "config.h"
#include "external/LinkedList/LinkedList.h"
#include "external/commonstructs.h"
#include "npc.h"
#include "config.h"

/* changelog
	- Added townwp and otherwp commands (ackmed)
	- Updated the MINOR_VERSION to reflect the real version (ackmed)
	- Added savepath/loadpath (Blackpower)
    - Added d2hackit bug/feature workaround that can cause a crash on game exit (ackmed)
	- Added autoBuyStop to global config (Blackpower)
	- Added textpatch support for magicprefix and suffix. (ackmed)
	- Made the last step of the NPC -> WP path be the location of the waypoint (ackmed)
	- Added steptime global config option. (ackmed)
    - Renamed intown -> townwp and outtown -> otherwp in the config file (ackmed)
    - Removed config sections size limit (ackmed)
	- Removed clear command.  any previously recorded path will be cleared when starting a new recording (ackmed)
	- Added intown=<string> & outtown=<string> to global config for getting waypoints (ackmed)
    - Added a bunch of WP stuff (ackmed)
    - Fixed typo on new setnpc (Blackpower)
	- Added Change to not record the same step twice when recording path (Blackpower)
	- Added globalconfig (ackmed)
	- Added autobuy (berto)
	- Added Client walk update (Blackpower)
	- Added Automaticly get NPC ID (Blackpower)
	- Changed setnpc to fit auto get npc id (Blackpower) 
	- Added name in NPCSTRUCT (Blackpower)              
	- Removed 0x13 capture since it's no longer needed (Blackpower)
	- Added support for bows (ackmed)
	- Added BaseName=<string> to config file (ackmed)
	- Added itemcode lookups to get items real name (ackmed)

*/
char ConfigPath[_MAX_PATH];  // dir where to find ini file
FUNCTIONENTRYPOINTS	*server;
WORD Debug;
char DebugBuffer[1024];
WORD state;
LinkedList *path = new LinkedList;
LinkedList *config = new LinkedList;


NPCSTRUCT	npc;
WPSTRUCT	wp;
THISGAMESTRUCT *thisgame;
CONFIGSTRUCTGLOBAL globalConfig;

BOOL Recording = false;
BOOL SetNPC = false;
BOOL isRunning = false;
BOOL inGame = false; // d2hackit bug/feature work around
DWORD tick = 0;
LinkedItem_t *currentStep;


// module info
DWORD	ModuleVersion=MAKELONG(VERSION_MAJOR,VERSION_MINOR);
char	ModuleAuthor[]="ackmed";
char	ModuleWebsite[]="www.gotwalls.com";
char	ModuleDescription[]="shopbot for Diablo2 LOD";
char	ModuleEmail[]="ackmed@gotwalls.com";

// module commands
MODULECOMMANDSTRUCT ModuleCommands[]= {
	{	
		"debug",
		OnCommandDebug,
		"c4debugc0 [devel|packet] toggle debug mode.\n"
		"[devel] adds extra debugging\n"
		"[packet] adds packet level debugging"
	},
	{
		"record",
		OnCommandRecord,
		"c4recordc0 toggles starting and stoping of path recording"
	},
/*	{
		"clear",
		OnCommandClear,
		"clearc0 removes the recorded path"
	}, */
	{
		"reload",
		OnCommandReload,
		"c4reloadc0 reloads ini file"
	},
	{
		"setnpc",
		OnCommandSetNPC,
		"setnpcc0 used to set which npc to trade with.\n"
		"c4setnpcc0 act<num> <npc name>\n"
		"c4setnpcc0 list act<num>"
	},
	{
		"townwp",
		OnCommandTownWP,
		"townwpc0 used to set the town waypoint.\n"
		"c4townwpc0 <waypoint name>\n"
	},
	{
		"otherwp",
		OnCommandOtherWP,
		"otherwpc0 used to set the other waypoint.\n"
		"c4otherwpc0 <waypoint name>\n"
	},
	{
		"savepath",
		OnCommandSavePath,
		"savepathc0 <pathname>"
		"Saves path to shopbot\\<pathname>.shp"
	},
	{
		"loadpath",
		OnCommandLoadPath,
		"loadpathc0 <pathname>\n"
		"Load shopbot\\<pathname>.shp"
	},
	{
		"help",
		OnCommandHelp,
		"c4setnpcc0 help!?"
	},
	{
		"start",
		OnCommandStart,
		"c4startc0 start the bot"
	},
	{
		"stop",
		OnCommandStop,
		"c4stopc0 stop the bot"
	},
	{NULL}
};

// module command functions
BOOL PRIVATE OnCommandDebug(char **argv, int argc) {

	// no params passed, just do a simple toggle
	if(argc == 2){
		if(Debug) {
			server->GamePrintString("Debugging disabled");
			Debug = DEBUG_OFF;
		} else {
			server->GamePrintString("Debugging enabled");
			Debug = DEBUG_USER;
		}
		return true;
	} else if(argc == 3) {
		if(strcmpi(argv[2],"devel") == 0) {
			server->GamePrintString("Devel Debugging enabled");
			Debug = DEBUG_DEVEL;
			return true;
		} else if(strcmpi(argv[2],"packet") == 0) {
			server->GamePrintString("Packet Debugging enabled");
			Debug = DEBUG_PACKET;
			return true;
		}
	}
	return false;
}


BOOL PRIVATE OnCommandRecord(char **argv, int argc) {
	LinkedItem_t *step;
	if(Recording) {
		server->GamePrintString("Path Recording Stopped");
	} else {
		step = path->GetFirstItem();
		if(step != NULL) {
			server->GamePrintString("Clearing Previously recorded path");
			for(step = path->GetFirstItem();step != NULL;step = path->RemoveItem(step)){
				delete (LOCATIONSTRUCT*) step->lpData;
			}
		}
		server->GamePrintString("Path Recording Started");
	}
	// toggle it
	Recording = 1 - Recording;
	return true;
}


/*BOOL PRIVATE OnCommandClear(char **argv, int argc) {
	WORD sndid = atoi(argv[2]);
	sprintf(DebugBuffer,"sndID:%d",sndid);
	server->GamePrintString(DebugBuffer);
	SendSoundToClient(thisgame->player->PlayerID,sndid);
	SendSoundToServer(sndid);
	
	return true;
}*/
 
BOOL PRIVATE OnCommandReload(char **argv, int argc) {
	LinkedItem_t *cfg;

	// delete the old config
	for(cfg = config->GetFirstItem();cfg != NULL;cfg = config->RemoveItem(cfg)){
		delete (CONFIGSTRUCT*) cfg->lpData;
	}

	// load new config
	if(!LoadConfig()) {
		server->GamePrintString("Error loading Config");
	} else {
		server->GamePrintString("shopbot.ini reloaded");
	}
	return true;
}

BOOL PRIVATE OnCommandTownWP(char **argv, int argc) {

	char waypoint[255];
	WPSTRUCT *wp;
	if(argc <= 2) {
		return false;
	}
	
	strcpy(waypoint,argv[2]);
	for(int i = 3; i < argc; i++) {
		sprintf(waypoint,"%s %s",waypoint,argv[i]);
	}

	wp = getWPbyName(waypoint);
	if(wp != NULL && wp->isTown) {
		memcpy(&globalConfig.townWP,wp,sizeof(WPSTRUCT));
		sprintf(DebugBuffer,"townwp set to: %s",waypoint);
		server->GamePrintString(DebugBuffer);
	} else {
		server->GamePrintString("Invalid townwp");
	}
	return true;
}

BOOL PRIVATE OnCommandOtherWP(char **argv, int argc) {

	char waypoint[255];
	WPSTRUCT *wp;
	if(argc <= 2) {
		return false;
	}
	
	strcpy(waypoint,argv[2]);
	for(int i = 3; i < argc; i++) {
		sprintf(waypoint,"%s %s",waypoint,argv[i]);
	}

	wp = getWPbyName(waypoint);
	if(wp != NULL) {
		memcpy(&globalConfig.otherWP,wp,sizeof(WPSTRUCT));
		sprintf(DebugBuffer,"otherWP set to: %s",waypoint);
		server->GamePrintString(DebugBuffer);
	} else {
		server->GamePrintString("Invalid townwp");
	}
	return true;
}
BOOL PRIVATE OnCommandSetNPC(char **argv, int argc) {
	if(argc != 4) {
		return false;
	}

	if(strcmpi("list",argv[2]) == 0) {
		// list possible npcs
		if(strcmpi("act1",argv[3]) == 0) {
			ListNPC(ACT1);
		} else if(strcmpi("act2",argv[3]) == 0) {
			ListNPC(ACT2);
		} else if(strcmpi("act3",argv[3]) == 0) {
			ListNPC(ACT3);
		} else if(strcmpi("act4",argv[3]) == 0) {
			ListNPC(ACT4);
		} else if(strcmpi("act5",argv[3]) == 0) {
			ListNPC(ACT5);
		} else {
			server->GamePrintString("Invalid act");
		}
		return true;
	} else if(strcmpi("act1",argv[2]) == 0) {
		SetNPCLid(ACT1,argv[3]);
	} else if(strcmpi("act2",argv[2]) == 0) {
		SetNPCLid(ACT2,argv[3]);
	} else if(strcmpi("act3",argv[2]) == 0) {
		SetNPCLid(ACT3,argv[3]);
	} else if(strcmpi("act4",argv[2]) == 0) {
		SetNPCLid(ACT4,argv[3]);
	} else if(strcmpi("act5",argv[2]) == 0) {
		SetNPCLid(ACT5,argv[3]);
	} else {
		server->GamePrintString("Invalid act");
	}
	return true;
}

BOOL PRIVATE OnCommandStart(char **argv, int argc) {
	state = STATE_WPFROMTOWN;
	server->GamePrintString("Starting shopbot");
	isRunning = true;
	return true;
}

BOOL PRIVATE OnCommandStop(char **argv, int argc) {
	state = STATE_NONE;
	server->GamePrintString("Stopping shopbot");
	isRunning = false;
	return true;
}

BOOL PRIVATE OnCommandSavePath(char **argv, int argc) {
	if(argc != 3)
		return false;

	if(npc.lid == 0) {
		server->GamePrintString("You must set an npc first!");
		return false;
	}
	if(strlen(argv[2])+strlen(ConfigPath) > 1024) {
		server->GamePrintString("File name too long!");
		return false;
	}
	
	char filename[1024];
	sprintf(filename, "%s\\shopbot\\%s.sbp", ConfigPath, argv[2]);

	FILE *stream;
	if((stream = fopen(filename, "w")) == NULL)
	{
		sprintf(DebugBuffer, "Error creating %s", filename);
		server->GamePrintString(DebugBuffer);
		return true;
	}
	fprintf(stream, "#This is a shopbot path file do not modify!\n");
	fprintf(stream, "NPCID: %d\n", npc.lid);
	fprintf(stream, "TOWNWP: %s\n", globalConfig.townWP.name);
	fprintf(stream, "OTHERWP: %s\n", globalConfig.otherWP.name);
	
	currentStep = path->GetFirstItem();
	
	while(currentStep != NULL) {
		LOCATIONSTRUCT *locp;
		locp = (LOCATIONSTRUCT*)currentStep->lpData;
		currentStep = path->GetNextItem(currentStep);
		fprintf(stream, "STEP: %d,%d\n",locp->x, locp->y);
	}
	fclose(stream);
	
	sprintf(DebugBuffer,"Path Saved to shopbot\\%s.sbp",argv[2]);
	server->GamePrintString(DebugBuffer);

	return true;
}

BOOL PRIVATE OnCommandLoadPath(char **argv, int argc) {

	if(argc != 3)
		return false;

	WORD lid;
	WORD x,y;

	char filename[1024], line[255];
	FILE *stream;
	LinkedItem_t *step;	

	//Delete path if present.
	for(step = path->GetFirstItem();step != NULL;step = path->RemoveItem(step)){
		delete (LOCATIONSTRUCT*) step->lpData;
	}

	sprintf(filename, "%s\\shopbot\\%s.sbp", ConfigPath, argv[2]);
	
	if((stream = fopen(filename, "r")) == NULL) {
		sprintf(DebugBuffer, "Unable to read from %s", filename);
		server->GamePrintString(DebugBuffer);
		return true;
	}

	// skip initial line
	while(fgetc(stream) != '\n') {};

	fscanf(stream, "NPCID: %u\n", &lid);

	if(Debug == DEBUG_DEVEL) {
			sprintf(DebugBuffer,"NPCID: %d",lid);
			server->GamePrintString(DebugBuffer);
	}

	for(int i = 0;i < NPC_TOTAL;i++)
	{
		if(allNPC[i].lid == lid)
		{
			npc = allNPC[i];
			SetNPC = true;
		}
	}

	if(!SetNPC) {
		server->GamePrintString("Unable to determine NPC from NPCID!");
		fclose(stream);
		return true;
	}

	
	WPSTRUCT *wp;
	// this is messed up.  I can't fscanf(stream,"TOWNWP: %[a-zA-Z0-9 ]s\n", line);
	fscanf(stream, "%s\n", line);
	fscanf(stream, "%[a-zA-Z0-9 ]s\n", line);

	if(Debug == DEBUG_DEVEL) {
		sprintf(DebugBuffer,"TOWNWP: %s",line);
		server->GamePrintString(DebugBuffer);
	}

	wp = getWPbyName(line);
	if(wp != NULL || !wp->isTown) {
		memcpy(&globalConfig.townWP,wp,sizeof(WPSTRUCT));
	} else {
		server->GamePrintString("Invalid town wp from path file.");
		fclose(stream);
		return true;
	}	
	
	// same deal.  can't fscanf(stream,"OTHERWP: %[a-zA-Z0-9 ]s\n", line);
	fscanf(stream, "%s\n", line);
	fscanf(stream, "%[a-zA-Z0-9 ]s\n", line);

	if(Debug == DEBUG_DEVEL) {
			sprintf(DebugBuffer,"OTHERWP: %s",line);
			server->GamePrintString(DebugBuffer);
	}

	wp = getWPbyName(line);
	if(wp != NULL) {
		memcpy(&globalConfig.otherWP,wp,sizeof(WPSTRUCT));
	} else {
		server->GamePrintString("Invalid other wp from path file.");
		fclose(stream);
		return true;
	}
		
	while(fscanf(stream,"%s",line) != EOF) {
		fscanf(stream,"%u,%u",&x, &y);
		LOCATIONSTRUCT *lstep = new LOCATIONSTRUCT;
		lstep->x = x;
		lstep->y = y;
		if(Debug == DEBUG_DEVEL) {
			sprintf(DebugBuffer,"STEP: %d %d",x,y);
			server->GamePrintString(DebugBuffer);
		}
		path->AddItem(lstep);
	}
	
	fclose(stream);

    sprintf(DebugBuffer, "Path Loaded and NPC set to %s", npc.name);
	server->GamePrintString(DebugBuffer);

	sprintf(DebugBuffer, "Go stand on the %s waypoint and run '.shopbot start'",globalConfig.townWP.name);
	server->GamePrintString(DebugBuffer);

	return true;
}

VOID EXPORT OnGameJoin(THISGAMESTRUCT* game) {
	LinkedItem_t *step;
	thisgame = game;
	// delete paths that may already exist
	for(step = path->GetFirstItem();step != NULL;step = path->RemoveItem(step)){
		delete (LOCATIONSTRUCT*) step->lpData;
	}
	inGame = true; // d2hackit bug/feature work around
}

VOID EXPORT OnGameLeave(THISGAMESTRUCT* game) {
	inGame = false; // d2hackit bug/feature work around
}

// functions called by d2hackit
BOOL EXPORT OnClientStart() {
	return LoadConfig();
}
// outgoing packet
DWORD EXPORT OnGamePacketBeforeSent(BYTE* aPacket, DWORD aLen) {
	// dump packet if packet debugging is on
	if(Debug == DEBUG_PACKET) {
		sprintf(DebugBuffer,"%s:","SENT: ");
		for(int i = 0;i < aLen;i++) {
			sprintf(DebugBuffer,"%s %.2x",DebugBuffer,aPacket[i]);
		}
		server->GamePrintInfo(DebugBuffer);
	}

	if(isRunning && aPacket[0] == 0x03) {
		//Client walk update, I think this is Icky's findings, not sure, but if it's
        //he should have the thanks for it:)

		// this needs to move into pachethelper.cpp
		BYTE upd[16];
        upd[0] = 0x0f;
        upd[6] = 0x17;
		memcpy (upd + 2, &thisgame->player->PlayerID, 4);
		memcpy (upd + 7, aPacket + 1, 4);
		memcpy (upd + 12, &thisgame->player->PlayerPositionX, 2);
		memcpy (upd + 14, &thisgame->player->PlayerPositionY, 2);
        server->GameSendPacketToGame (upd, 16);
        return aLen;
    }

	

	// if recording save any movement 
	if(Recording) {
		if(aPacket[0] == 0x03) {
			LOCATIONSTRUCT *step = new LOCATIONSTRUCT;
			LOCATIONSTRUCT *prevStep;
			memcpy(&step->x,&aPacket[1],2);
			memcpy(&step->y,&aPacket[3],2);
			if(path->GetLastItem() != NULL) {
				prevStep = (LOCATIONSTRUCT*)path->GetLastItem()->lpData;
				// only record if its a new step
				if(step != prevStep) {
					if(Debug == DEBUG_DEVEL) {
						sprintf(DebugBuffer,"Recording Step: %d, %d",step->x,step->y);
						server->GamePrintString(DebugBuffer);
					}
					path->AddItem(step);
				} else {
					delete step;
				}
			} else {
				path->AddItem(step);
			}
		}
	}

	// deny client sync with server while bot is running
	if(isRunning && aPacket[0] == 0x5f) {
		if(Debug == DEBUG_DEVEL) {
			server->GamePrintString("Denying position sync");
		}
		return 0;
	}
	return aLen;

}

// incoming packet
DWORD EXPORT OnGamePacketBeforeReceived(BYTE* aPacket, DWORD aLen) {

	// dump packet if packet debugging is on
	if(Debug == DEBUG_PACKET) {
		sprintf(DebugBuffer,"%s","RECV: ");
		for(int i = 0;i < aLen;i++) {
			sprintf(DebugBuffer,"%s %.2x",DebugBuffer,aPacket[i]);
		}
		server->GamePrintInfo(DebugBuffer);
	}

/*	if(state != STATE_NONE && (aPacket[0] == 0x63) ) {
		// work around to send empty packet taken from
		// spoof module by Sonata.  THX!
		memset(aPacket, 0, aLen);	// Zero out the packet
		aPacket[0] = 0x77;			// Turn it into a dummy
		aPacket[1] = 0x06;
		return aLen;
	}

	*/

	if(SetNPC && aPacket[0] == 0x55) {
		WORD check;
		memcpy( &check, &aPacket[5], 2);
        if(npc.lid == check) {
			memcpy(&npc.npcID, &aPacket[1], 4);
			if(Debug == DEBUG_DEVEL) {
				sprintf(DebugBuffer, "Got %s ID(%d)", npc.name, npc.npcID);
                server->GamePrintInfo(DebugBuffer);
			}
            SetNPC = false;
		}
	}
	if(aPacket[0] == 0x96) { // location sync from server
		if(Debug == DEBUG_DEVEL) {
			BitFields iPacket(aPacket,aLen);
			iPacket.GetField(23);
			LOCATIONSTRUCT loc;
			loc.x = iPacket.GetField(16);
			loc.y = iPacket.GetField(16);
			sprintf(DebugBuffer,"Location Sync from server (%d,%d)",loc.x,loc.y);
			server->GamePrintString(DebugBuffer);
		}
	} else if(aPacket[0] == ITEM_MESSAGEID_DROP) {
		ITEMSTRUCT *item;
		item = ParseItemDrop(aPacket,aLen);
		if((isRunning || Debug == DEBUG_DEVEL) && item != NULL) {
			DumpItemStruct(item);
		}
		// check for match needs to go here
		if(item != NULL) {
			CONFIGSTRUCT *cfg;
			cfg = ItemMatchCheck(item);
			if(cfg != NULL) {
				// we have a match
				
				// print stuff if description is defined.
				if(cfg->Description[0] != '\0') {
					server->GamePrintString(cfg->Description);
				}
				if(globalConfig.autoBuy) {
					server->GamePrintString("AutoBuy Enabled: Attempting to purchase item");
					BuyItem(item->ItemID, npc.npcID);
					if(globalConfig.autoBuyStop) {
						server->GamePrintString("Bot going into anti-idle mode.  Please run '.shopbot stop' to fully stop him");
						state = STATE_ANTIIDLE;
						isRunning = false;
					} else {
						server->GamePrintString("Bot is continuing to shop, autoBuyStop is disabled");
					}
				} else {
					server->GamePrintString("Bot going into anti-idle mode.  Please run '.shopbot stop' to fully stop him");
					state = STATE_ANTIIDLE;
					isRunning = false;
				} 
			} 
			delete item;
		}
	} else if(aPacket[0] == 0x6d) {  // npc update
		if(npc.npcID != 0) {
			WORD npcID;
			memcpy(&npcID,&aPacket[1],4);
			// see if this is our npc
			if(npc.npcID == npcID) {
				memcpy(&npc.pos.x,&aPacket[5],2);
				memcpy(&npc.pos.y,&aPacket[7],2);
				if(Debug == DEBUG_DEVEL) {
					sprintf(DebugBuffer,"Got NPC loc update (%d,%d)",npc.pos.x,npc.pos.y);
					server->GamePrintString(DebugBuffer);
				}
			}
		}
	} else if(aPacket[0] == 0x51) {	// possible waypoint info
		if(aPacket[1] == 0x02) { 
			DWORD lid;
			memcpy(&lid,&aPacket[6],2);
			if(isWP(lid)) {
				memcpy(&wp.pos.x,&aPacket[8],2);
				memcpy(&wp.pos.y,&aPacket[10],2);
				memcpy(&wp.entityID,&aPacket[2],4);
				wp.lid = lid;
				// should fill in position at some point.
				if(Debug == DEBUG_DEVEL) {
					sprintf(DebugBuffer,"wp id: %d (%d,%d) lid:%.4x",wp.entityID,wp.pos.x,wp.pos.y,wp.lid);
					server->GamePrintString(DebugBuffer);
				}
				wp.isStale = false;
			}
		} 
	}

	return aLen;

}

// is run ~every 1/10th a second
DWORD EXPORT OnGameTimerTick(void) {
	
	// d2hackit bug/feature work around
	// OnGameTimerTick() is called even while outside of a game
	// this can lead to diablo crashing.
	if(!inGame) {
		return true;
	}

	tick++;
	if(state == STATE_NONE) {
		tick = 0;
	} else if(tick > 100) { // if we have been in a state for more then 10 seconds something is wrong
		server->GamePrintString("State: Lost.  Attempting to retake path to WP");
		state = STATE_WALKWP_START;
		wp.isStale = false;
		tick = 0;
	} else {
		switch(state) {
			case STATE_ANTIIDLE:
				if(tick == 20) {
					TakeStep(thisgame->player->PlayerPositionX+1,thisgame->player->PlayerPositionY + 1);
				} else if(tick == 40) {
					TakeStep(thisgame->player->PlayerPositionX-1,thisgame->player->PlayerPositionY - 1);
				} else if(tick > 60) {
					SendSoundToServer(26);  // this is for you;)
					tick = 0;
				}
				break;
			case STATE_WPFROMTOWN:	// leave town.
				if(/*tick > 3 && */!wp.isStale) {
					server->GamePrintString("State: Taking WP out of town");
					TakeWayPoint(wp.entityID,globalConfig.otherWP.wpNum);
				
					// next state is taking wp back to town.
					state = STATE_WPTOTOWN;
					wp.isStale = true;
					tick = 0; // reset tick
				}
				break;
			case STATE_WPTOTOWN:
				if(!wp.isStale) {
					server->GamePrintString("State: Taking WP to town");
					TakeWayPoint(wp.entityID,globalConfig.townWP.wpNum);
			
					// next state is to start walking to npc
					state = STATE_WALKNPC_START;
					tick = 0;
				}
				
				break;
			case STATE_WALKNPC_START:
				if(npc.npcID != 0) {
					server->GamePrintString("State: Walking to NPC");
					currentStep = path->GetFirstItem();
					// start walking
					state = STATE_WALKNPC;
					tick = 0;
				} else {
					// dont have an npc!?
					server->GamePrintString("Dont have an NPC!, shopbot disabled");
					state = STATE_NONE;
				}
				break;
			case STATE_WALKNPC:
				if(currentStep != NULL) {
					if(tick > globalConfig.steptime) { // dont try walking too fast, should make this adjustable
						LOCATIONSTRUCT *loc;
						loc = (LOCATIONSTRUCT*)currentStep->lpData;
						if(Debug == DEBUG_DEVEL) {
						//	ForceClientViewLocation(loc->x,loc->y,0);
					//		server->GamePrintString("Taking Step to NPC");
						}
						TakeStep(loc->x,loc->y);
						currentStep = path->GetNextItem(currentStep);
						
						// last point in our recorded path, try talking to where the NPC is located
						if(currentStep == NULL) {
							if(Debug == DEBUG_DEVEL) {
								server->GamePrintString("Taking Last Step to NPC");
							}
							TakeStep(npc.pos.x,npc.pos.y);
							state = STATE_OPENSTORE;
						}
						// reset tick
						tick = 0;
					}		
				}
				break;
			case STATE_OPENSTORE:
				if(tick > 3) {
					server->GamePrintString("State: Opening Store with NPC");
					OpenStore(npc.npcID,npc.pos.x,npc.pos.y);
					state = STATE_CLOSESTORE;
					tick = 0;
				}
				break;
			case STATE_CLOSESTORE:
				if(tick > 7) { // wait ~7/10th second before closing store to allow packets to get here
					server->GamePrintString("State: Closing Store with NPC");
					CloseStore(npc.npcID);
					state = STATE_WALKWP_START;
					tick = 0;
				}
				break;
			case STATE_WALKWP_START:  // basicly the same as walking to npc, but we traverse the linkedlist backwords
				if(wp.entityID != 0) {
					server->GamePrintString("State: Walking to WP");
					currentStep = path->GetLastItem();
					// start walking
					state = STATE_WALKWP;
					tick = 0;
				} else {
					// dont have an wp!?
					server->GamePrintString("Dont have an WP!, shopbot disabled");
					state = STATE_NONE;
				}
				break;
			case STATE_WALKWP:
				if(currentStep != NULL) {
					if(tick > globalConfig.steptime) { // dont try walking too fast, should make this adjustable
						LOCATIONSTRUCT *loc;
						loc = (LOCATIONSTRUCT*)currentStep->lpData;
						if(Debug == DEBUG_DEVEL) {
						//	ForceClientViewLocation(loc->x,loc->y,0);
						// server->GamePrintString("Taking Step to WP");
						}
						TakeStep(loc->x,loc->y);
						currentStep = path->GetPrevItem(currentStep);
						
						// first point in our recorded path 
						if(currentStep == NULL) {
							if(Debug == DEBUG_DEVEL) {
								server->GamePrintString("Taking Last Step to WP");
							}
							TakeStep(wp.pos.x,wp.pos.y);
							state = STATE_WPFROMTOWN;
						}
						// reset tick
						tick = 0;
					}		
				}
				break;
			default:
				break;
		}
	}
	return true;
}

BOOL LoadConfig(void) {
	char ConfigFile[1024];
	char *Sections;
	char error[1024];
	char *rval;
	int length,counter;
	CONFIGSTRUCT *cfg;

	sprintf(ConfigFile,"%s\\%s",ConfigPath,"shopbot.ini");

	if (_access(ConfigFile, 0)) { 
		char error[1024];
		sprintf(error, "Unable to open Config: %s", ConfigFile);
		server->GamePrintError(error);
		return false;
	}

	// Try getting the data in 64-byte increments
	DWORD  alloc=0;
	DWORD  allocStep = 20;

	while (TRUE) {
		alloc+=allocStep;
		Sections = new char[alloc];
		length = GetPrivateProfileSectionNames(Sections,alloc,ConfigFile);
		
		if(length + 1 != alloc - 1 ) {
			break;
		}
		delete Sections;
	}

	counter = 0;
NEXT:
	while(counter < length) {
		cfg = new CONFIGSTRUCT;
		cfg->Description[0] = '\0';
		cfg->BaseName[0] = '\0';
		cfg->MagicPrefix[0] = '\0';
		cfg->MagicSuffix[0] = '\0';
		cfg->item.ItemCode[0] = '\0';

		if(strlen(&Sections[counter]) > 128 || Sections[counter] == '\0') {
			
			if(Debug) {
				sprintf(DebugBuffer,"LoadConfig: Bad Section length (%d)",strlen(&Sections[counter]));
				server->GamePrintInfo(DebugBuffer);
			}
			counter = counter + strlen(&Sections[counter]) + 1;
			delete cfg;
			goto NEXT;
		}
		
		
		if(strcmpi("global",&Sections[counter]) == 0) {
			if(Debug == DEBUG_DEVEL) {
				server->GamePrintString("Loading Global Config Section");
			}
			counter = counter + strlen(&Sections[counter]) + 1;
			delete cfg;

			rval = server->GetHackProfileString("shopbot","global","autobuy");
			if(strlen(rval)) {
				if(atoi(rval)) {
					globalConfig.autoBuy = true;
				} else {
					globalConfig.autoBuy = false;
				}
			} else {
				globalConfig.autoBuy = false;
			}

			rval = server->GetHackProfileString("shopbot","global","autobuystop");
			if(strlen(rval)) {
				if(atoi(rval)) {
					globalConfig.autoBuyStop = true;
				} else {
					globalConfig.autoBuyStop = false;
				}
			} else {
				globalConfig.autoBuyStop = true;
			}
			
			WPSTRUCT *wp;
			rval = server->GetHackProfileString("shopbot","global","townwp");
			if(strlen(rval)) {
				wp = getWPbyName(rval);
				if(wp == NULL) {
					sprintf(DebugBuffer,"Invalid townwp waypoint name '%s'",rval);
					server->GamePrintString(DebugBuffer);
					globalConfig.townWP.wpNum = 0;
				} else {
					if(!wp->isTown) {
						server->GamePrintString("townwp waypoint must be a town!");
						globalConfig.townWP.wpNum = 0;
					} else {
						memcpy(&globalConfig.townWP,wp,sizeof(WPSTRUCT));
					}
				}
			} else {
				globalConfig.townWP.wpNum = 0;
			}

			rval = server->GetHackProfileString("shopbot","global","otherwp");
			if(strlen(rval)) {
				wp = getWPbyName(rval);
				if(wp == NULL) {
					sprintf(DebugBuffer,"Invalid otherwp waypoint name '%s'",rval);
					server->GamePrintString(DebugBuffer);
					globalConfig.otherWP.wpNum = 0;
				} else {
						memcpy(&globalConfig.otherWP,wp,sizeof(WPSTRUCT));
				}
			} else {
				globalConfig.otherWP.wpNum = 0;
			}

			rval = server->GetHackProfileString("shopbot","global","steptime");
			if(strlen(rval)) {
				globalConfig.steptime = atoi(rval);
			} else {
				globalConfig.steptime = 2;
			}

			
			if(globalConfig.townWP.wpNum == 0 || globalConfig.otherWP.wpNum == 0) {
				server->GamePrintString("You have invalid or non-existing intown/outtown global settings");
				return false;
			}
			if(Debug) {
				sprintf(DebugBuffer,"[global] townWP: %s otherWP: %s autobuy %d steptime: %d",globalConfig.townWP.name,globalConfig.otherWP.name,globalConfig.autoBuy,globalConfig.steptime);
				server->GamePrintString(DebugBuffer);
			}
			goto NEXT;
		} 
		
		strcpy(cfg->Section,&Sections[counter]);
		counter = counter + strlen(&Sections[counter]) + 1;
		// now query the section and fill in data
		
		// ItemCode should be 3 or 4 chars long
		rval = server->GetHackProfileString("shopbot", cfg->Section, "Code");
		if(strlen(rval)) {
			if(strlen(rval) != 4 && strlen(rval) != 3) {
				sprintf(error,"Bad Code value '%s' in section [%s], Skipping",rval,cfg->Section);
				server->GamePrintString(error);
				delete cfg;
				goto NEXT;
			}
			
			// if 4 char is a space, chop it off
			if(rval[3] == ' ') 
				rval[3] = '\0';

			strcpy(cfg->item.ItemCode,rval);
			// delete rval;
		} else {
			cfg->item.ItemCode[0] = '\0';
		}

		// Item Description
		rval = server->GetHackProfileString("shopbot", cfg->Section, "Description");
		if(strlen(rval)) {
			if(strlen(rval) > 127) {
				sprintf(error,"Description too long in section [%s], Skipping",cfg->Section);
				server->GamePrintString(error);
				delete cfg;
				goto NEXT;
			}
			strcpy(cfg->Description,rval);
			// delete rval;
		} else {
			cfg->Description[0] = '\0';
		}

		// basename of item, aka itemname without a prefix or suffix
		rval = server->GetHackProfileString("shopbot", cfg->Section, "BaseName");
		if(strlen(rval)) {
			if(strlen(rval) > 127) {
				sprintf(error,"BaseName too long in section [%s], Skipping",cfg->Section);
				server->GamePrintString(error);
				delete cfg;
				goto NEXT;
			}
			strcpy(cfg->BaseName,rval);
			// delete rval;
		} else {
			cfg->BaseName[0] = '\0';
		}

		rval = server->GetHackProfileString("shopbot", cfg->Section, "MagicPrefix");
		if(strlen(rval)) {
			if(strlen(rval) > 127) {
				sprintf(error,"MagicPrefix too long in section [%s], Skipping",cfg->Section);
				server->GamePrintString(error);
				delete cfg;
				goto NEXT;
			}
			strcpy(cfg->MagicPrefix,rval);
			// delete rval;
		} else {
			cfg->MagicPrefix[0] = '\0';
		}

		rval = server->GetHackProfileString("shopbot", cfg->Section, "MagicSuffix");
		if(strlen(rval)) {
			if(strlen(rval) > 127) {
				sprintf(error,"MagicSuffix too long in section [%s], Skipping",cfg->Section);
				server->GamePrintString(error);
				delete cfg;
				goto NEXT;
			}
			strcpy(cfg->MagicSuffix,rval);
			 // delete rval;
		} else {
			cfg->MagicSuffix[0] = '\0';
		}
		// add to config linked list.
		config->AddItem(cfg);
		if(Debug == DEBUG_DEVEL) {
			sprintf(DebugBuffer,"[%s] des='%s' pre='%s' code='%s' basename='%s' suf='%s'",cfg->Section,cfg->Description,cfg->MagicPrefix,cfg->item.ItemCode,cfg->BaseName,cfg->MagicSuffix);
			server->GamePrintString(DebugBuffer);
		}
	}
	return true;
}

CONFIGSTRUCT * ItemMatchCheck(ITEMSTRUCT *item) {
	LinkedItem_t *cfgItem;
	CONFIGSTRUCT *cfg;

	BOOL match;
	for(cfgItem = config->GetFirstItem();cfgItem != NULL;cfgItem = config->GetNextItem(cfgItem)){
		match = true;
		cfg = (CONFIGSTRUCT*)cfgItem->lpData;

		if(cfg->item.ItemCode[0] != '\0' && strcmpi(cfg->item.ItemCode,item->ItemCode) != 0) 
			match = false;

		if(cfg->MagicPrefix[0] != '\0') {
			if(item->PrefixName[0] != '\0') {
				if(strcmpi(cfg->MagicPrefix,item->PrefixName) != 0) {
					match = false;
				}
			} else {
				match = false;
			}
		}
		
		if(cfg->BaseName[0] != '\0') {
			if(item->BaseName[0] != '\0') {
				if(strcmpi(cfg->BaseName,item->BaseName) != 0) {
					match = false;
				}
			} else {
				match = false;
			}
		}

		if(cfg->MagicSuffix[0] != '\0') {
			if(item->SuffixName[0] != '\0') {
				if(strcmpi(cfg->MagicSuffix,item->SuffixName) != 0) {
					match = false;
				}
			} else {
				match = false;
			}
		}
		if(match) {
			return cfg;
		}

	}
	return NULL;
}




// 
// stuff below here is from ClientCore.cpp
//////////////////////////////////////////////////////////////////////
// Dll entry/exit
//////////////////////////////////////////////////////////////////////
BOOL APIENTRY DllMain(HANDLE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved)
{
	BOOL hResult = TRUE;
	char *t;
    switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
			
			// Create server struct			
			server = new FUNCTIONENTRYPOINTS; 

			// Bind exported functions from server
			HMODULE hModule;
			hModule = (HMODULE)GetModuleHandle("D2HackIt");
			
			// Macros make this look a lot nicer :)
			BIND_TO_SERVER(GamePrintInfo);
			BIND_TO_SERVER(GamePrintError);
			BIND_TO_SERVER(GamePrintVerbose);
			BIND_TO_SERVER(GamePrintString);
			BIND_TO_SERVER(GameCommandLine);
			BIND_TO_SERVER(GameSendPacketToServer);
			BIND_TO_SERVER(GameSendPacketToGame);
			BIND_TO_SERVER(GetFingerprint);
			BIND_TO_SERVER(Intercept);
			BIND_TO_SERVER(GetHackProfileString);
			BIND_TO_SERVER(SetHackProfileString);
			BIND_TO_SERVER(GetThisgameStruct);

			// Get plugin path
			t = ConfigPath;
			if (GetModuleFileName((HINSTANCE)hModule, t, _MAX_PATH)) {
				int p=strlen(ConfigPath);
				while (p) {
						if (ConfigPath[p] == '\\')
							{ ConfigPath[p] = 0; p=0;}
						else
					p--;
				}
			}	 
			// initiate client
			// hResult = OnClientStart();
			break;

		case DLL_PROCESS_DETACH:
			// kill client

			// hResult = OnClientStop();

			delete server;
			break;
    } 
    return hResult;
}

//////////////////////////////////////////////////////////////////////
// Stubfunctions for 'property get' functions.
//////////////////////////////////////////////////////////////////////
LPCSTR	EXPORT GetModuleAuthor()		{return ModuleAuthor;}
LPCSTR	EXPORT GetModuleWebsite()		{return ModuleWebsite;}
DWORD	EXPORT GetModuleVersion()		{return ModuleVersion;}
LPCSTR	EXPORT GetModuleEmail()			{return ModuleEmail;}
LPCSTR	EXPORT GetModuleDescription()	{return ModuleDescription;}

//////////////////////////////////////////////////////////////////////
// OnClientCommandLine
// -------------------------------------------------------------------
// The modules own extension of the command line interface. Any custom
// commands you add are parsed here.
//
// Return value should be TRUE, but it is not used at this 
// time.
//
// Arguments when we get here:
// argv[0]			Name of module
// argv[1]			Name of command (If supplied)
// argv[2 ... n]	The rest
//
// Syntax in the game: .<module> <arguments>
//////////////////////////////////////////////////////////////////////
BOOL EXPORT OnGameCommandLine(char* argv[], int argc)
{
	// Check if user supplied anything at all, if not assume help...
	if (argc==1)
		argv[argc++]="help";


	MODULECOMMANDSTRUCT* mcs=ModuleCommands;

	while (mcs->szName) {
		if (!stricmp(mcs->szName, argv[1]))
			break;
		mcs++;
	}

	char *p,*t,*msg,*fMsg;
	fMsg=new char[256];
	//
	// Is this a built-in function ?
	if (mcs->szName) {
		//
		// If functions returns false, show usage help
		if (!mcs->pFunc(argv, argc)) {
			t=new char[strlen(mcs->szUsage)+1];
			server->GamePrintInfo("Usage:");
			sprintf((char*)t, "%s", mcs->szUsage);
			if (strlen((char*)t))
			{
				msg=p=t;
				while (*p != 0) {
					if (*p == '\n') 
					{
						*(p++) = 0;
						sprintf(fMsg, "c4.%s %s", argv[0], msg);
						server->GamePrintInfo((char*)fMsg);
					if (*p != 0)
						msg = p;
					} else
						p++;
					}
				sprintf(fMsg, "c4.%s %s", argv[0], msg);
				server->GamePrintInfo((char*)fMsg);
			}
			delete t;
		}
	} else {
	// Unknown command, show catch-all help phraze.
	t=new char[128];
	sprintf(t, "Unknown command c4'%s %s'c0 - try c4'.%s help'c0 to get help.",
		argv[0], argv[1], argv[0]);
	server->GamePrintError(t);
	delete t;
	}
	delete fMsg;
	return TRUE;
}


//////////////////////////////////////////////////////////////////////
// OnCommandHelp
// -------------------------------------------------------------------
// Our default help function.
//
// Syntax in the game: .<module> <arguments>
//////////////////////////////////////////////////////////////////////
BOOL PRIVATE OnCommandHelp(char** argv, int argc)
{
	// If command line is longer than 2, show usage for 'help'
	if (argc>2) return FALSE;

	char t[1024];
	sprintf(t, "Available commands for %s:", argv[0]);
	server->GamePrintInfo(t);

	// Loop through ModuleCommands[] and print their names
	for (int i=0; ModuleCommands[i].szName != NULL; i++)
	{
		sprintf(t, "c4.%s %s", argv[0], ModuleCommands[i].szName);
		server->GamePrintInfo(t);
	}

	sprintf(t, "For help on a specific command, type c4.%s <command> help", argv[0]);
	server->GamePrintInfo(t);
	return TRUE;
}
